

set -ex



gdal_grid --version
gdal_rasterize --version
gdal_translate --version
gdaladdo --version
gdalenhance --version
gdalwarp --version
gdalinfo --formats
gdalinfo http://thredds.nersc.no/thredds/dodsC/greenpath/Model/topaz
bash /Users/builder/fwaters/gdal/gdal-feedstock/recipe/run_test.sh
exit 0
